/***************************************************************************/
/*                                                                         */
/*  Logger.c                                                               */
/*             Datalogger fr Fronius Wechselrichter mit LCD Anzeige       */
/*             PIC16F877  4Mhz                                             */
/*                                                                         */
/*      Mode : HD44780 type LCD displays                                   */
/*                                                                         */
/* Interface :     SEL = Port D 4                                          */
/*                  RS = Port D 5                                          */
/*              Data_7 = Port D 3                                          */
/*              Data_6 = Port D 2                                          */
/*              Data_5 = Port D 1                                          */
/*              Data_4 = Port D 0                                          */
/*                 SDA = Port C 4                                          */
/*                 SCL = Port C 3                                          */
/*                                                                         */
/* (c) Matthias Stehr 09.02.2002                                           */
/*                                                                         */
/*                                                                         */
/***************************************************************************/
#include <16F877.h>
#use delay(clock=4000000)          // Crystal on the board
#use i2c(MASTER, SDA=PIN_C4, SCL=PIN_C3)
#bit TX9D = 0x98.0 //the 9th bit for hardware uart tx
#include <lcd_d.c>    // LCD an Port D

#define INIT 0x11
#define ACPB 0x01
#define SUB  0x02
#define NUB  0x03
#define TB   0x04
#define WRSB 0x05
#define WW   0x06
#define FB   0xFF

#define GETDATA   0x01
#define SETTIME   0x02
#define RESET     0x03
#define RESTART   0x04


#define PCBAUD 3200

char timed_getc(void) ;
void get_data_wr(void);
void send_data_to_air(void);
void send_data_to_pc(void);
void display_data(void);
void calculate_data(void);
void read_PC_Command(void);
void ReadTime(void);
void WriteTime(void);
char BcdToChar(char c);
char CharToBcd(char b);
void read_EE(void);
void clear_EE(void);



unsigned int mb,acp,su,nu,t,cs,i,csair;  // 8 Bit  unsigned
unsigned int wrstatus;
long p,tt,u;                             // 16 Bit unsigned
short timeout_error, wr_error, cmdmode;
float daypower;
long power;
long w;
unsigned int wh,wl;
unsigned int ss, hh, oldhh,mm, day, month;
unsigned int index;

#define MAXCMD 10
#define BUFFER_SIZE 32
byte buffer[BUFFER_SIZE];
byte next_in = 0;
byte next_out = 0;


#use rs232(baud=PCBAUD,xmit=PIN_B1, rcv=PIN_B0, BITS=8)
#int_ext
rb0_isr()
{
	int t;

  buffer[next_in]=getc();
  t=next_in;
  next_in=(next_in+1) % BUFFER_SIZE;
  if(next_in==next_out)
  	next_in=t;           // Buffer full !!
}


#define bkbhit (next_in!=next_out)

byte bgetc() {
   byte c;

   //while(!bkbhit) ;
   c=buffer[next_out];
   next_out=(next_out+1) % BUFFER_SIZE;
   return(c);
}



/**************************************************************************/
/*                                                                        */
/**************************************************************************/
void main()
{
   ext_int_edge(H_TO_L);
   //enable_interrupts(int_ext);
   enable_interrupts(global);
   output_low(PIN_E2);  // deactivate sender
   TX9D=0;              // Bit 9th is zero and never used, because no data 
                        // are sent to the inverter
   daypower = 0;
   acp      = 0;
   su       = 0;
   nu       = 0;
   t        = 0;
   wl       = 0;
   wh       = 0;
   
   ss       = 0;
   mm       = 0;
   hh       = 0;
   oldhh    = hh;
   index    = 2;
   cmdmode  = false;
   
   WriteTime();
   lcd_init();
   printf(lcd_putc,"\fInit ...");
   delay_ms(1000);
   do
   {
      get_data_wr();
      if (bkbhit)
		  {
		    read_PC_Command();
		  }
		  ReadTime();
      if(!wr_error)
      {
        calculate_data();
        send_data_to_air();
        send_data_to_pc();
 		  	display_data();      		
      }
      else
      {
        printf(lcd_putc,"\fW=%6.0fWh",daypower);
        printf(lcd_putc,"\nWR ist AUS :-(");
        send_data_to_air();
      }
      // check if data must be stored in EE Prom
      if(hh != oldhh)
      {
      	if(hh == 0x00)
      	{
      		printf(lcd_putc,"\fSchreibe ins\nEE-Prom");
      		power = (long)daypower;
      		write_eeprom(index++,power);
      		write_eeprom(index++,power >> 8);
      		daypower = 0;
      		if(index == 0xFE)
      			index = 0x02;
      	}
      	oldhh = hh;
      }
   }
   while(TRUE);
}

/**************************************************************************/
/*                                                                        */
/**************************************************************************/
#use rs232(baud=3200, xmit = PIN_C6, rcv=PIN_C7, BITS=9, ERRORS)
char timed_getc() 
{ 
  long timeout;
	timeout_error=FALSE; 
	timeout=0; 
	while(!kbhit()&&(++timeout<10000)) // 10 second 
	{
	  delay_ms(1);
		if(!input(PIN_A0))      // PC Command Mode Switch activated ?
	  {
	     printf(lcd_putc,"\fPC COMMAND MODUS AKTIV");
	     cmdmode = true;
	     enable_interrupts(int_ext);
	  } 
	}
	if(kbhit()) 
		return(getc()); 
	else 
	{ 
		timeout_error=TRUE; 
		return(0); 
	}
}


/**************************************************************************/
/*                                                                        */
/**************************************************************************/
get_data_wr()
{
  wr_error = TRUE;
  do
  {
    mb  = timed_getc();
    if(timeout_error)
      return;
  }
  while(mb != INIT);    // Sync to CMD Byte
  acp = timed_getc();   // P = Byte*110/8
  if(timeout_error)
      return;
  su  = timed_getc();   // Solar U = Byte*4/3
  if(timeout_error)
      return;
  nu  = timed_getc();   // Netz U 
  if(timeout_error)
      return;
  t   = timed_getc();   // Temperatur T = (Byte -180)*3/2
  if(timeout_error)
      return;
  cs  = timed_getc();   // Checksum
  if(timeout_error)
      return;
  wr_error = FALSE;    
} 

/**************************************************************************/
/*                                                                        */
/**************************************************************************/
#use rs232(baud=1200,xmit=PIN_B2)
send_data_to_air()
{ 
 	output_high(PIN_E2);  // activate sender
 	wrstatus = wr_error;
 	w = daypower;
 	wl = w;
 	wh = w >> 8;
 	
 	/**************************************************************************/
 	// calculate new checksum for acp and temperature
  csair  = INIT;
  csair += ACPB;
  csair += acp;
  csair += t;
  // send data to air 2 times
  for(i=0; i<=2; i++)
  {
    putc(INIT);
 	  putc(ACPB);
		putc(acp);
 		putc(t);
 		putc(csair);
 	}
 	delay_ms(100);
 	/**************************************************************************/
 	// calculate new checksum for su and nu
  csair  = INIT;
  csair += SUB;
  csair += su;
  csair += nu;
  // send data to air 2 times
  for(i=0; i<=2; i++)
  {
    putc(INIT);
 		putc(SUB);
 		putc(su);
 		putc(nu);
 		putc(csair);
 	}
 	delay_ms(100);
 	/**************************************************************************/ 	
 	// calculate new checksum for wrstatus
  csair  = INIT;
  csair += WRSB;
  csair += wrstatus;
  csair += FB;
  // send data to air 2 times
  for(i=0; i<=2; i++)
  {
    putc(INIT);
 		putc(WRSB);
 		putc(wrstatus);
 		putc(FB);
 		putc(csair);
 	}
 	delay_ms(100);
 	/**************************************************************************/ 	
 	// calculate new checksum for W
  csair  = INIT;
  csair += WW;
  csair += wl;
  csair += wh;
  // send data to air 2 times
  for(i=0; i<=2; i++)
  {
    putc(INIT);
 		putc(WW);
 		putc(wl);
 		putc(wh);
 		putc(csair);
 	}
 	delay_ms(100);
 	output_low(PIN_E2);  // deactivate sender
}



/**************************************************************************/
/*                                                                        */
/**************************************************************************/
#use rs232(baud=PCBAUD,xmit=PIN_B1, rcv=PIN_B0, BITS=9)
send_data_to_pc()
{
  bit_set(RS232_ERRORS,7);   // Bit 7 is the 9th bit for 9 bit mode 
  //send original wr-data to PC
  putc(INIT);
 	putc(acp);
 	putc(su);
 	putc(nu);
 	putc(t);
 	putc(cs);
}


/**************************************************************************/
/*                                                                        */
/**************************************************************************/
display_data()
{ 
  unsigned int m,d;
  
  if (cmdmode)
  {
    printf(lcd_putc,"\fPC COMMAND MODUS AKTIV");
  }
  else
  {
    m = read_eeprom(0x00); // get start month
    d = read_eeprom(0x01); // get start day
	  printf(lcd_putc,"\fW=%6.0fWh Us=%3luV %02u/%02u",daypower ,u ,d,m);
    //printf(lcd_putc,"\fP=%04luW Us=%03luV",p,u);
    //printf(lcd_putc,"\nT=%03luC Un=%03uV",tt,nu); 
    //printf(lcd_putc,"\nW=%6.0fWh %02u:%02u:%02u",daypower,hh,mm,ss); 
    printf(lcd_putc,"\n%02u:%02u:%02u P=%4luW T=%03luC",hh,mm,ss, p, tt);
  }
}


/**************************************************************************/
/*                                                                        */
/**************************************************************************/
calculate_data()
{
  p   = (long)acp;
  p   = p * 110;
  p   = p / 8;
    
  u   = (long) su;
  u   = u * 4;
  u   = u / 3;
    
  tt  = (long) t;
  tt  = tt - 180;
  tt  = tt * 3;
  tt  = tt / 2;
  
  daypower += ((float)p*4.8)/3600.0;  
}


/**************************************************************************/
/*                                                                        */
/**************************************************************************/
#use rs232(baud=PCBAUD,xmit=PIN_B1, rcv=PIN_B0, BITS=8)
read_PC_Command()
{
	byte cc, cmd[MAXCMD]; 
	
	cc = 0;
  while(bkbhit && (cc < MAXCMD))      // read the buffer
 		cmd[cc++] = bgetc();	
 	if(cmd[0] != 0x00)    // check trigger
 	{
 	  disable_interrupts(int_ext);
 	  cmdmode = false;
 		return;
 	}
 	if(cc < 2)            // check min length
 	{
 	  disable_interrupts(int_ext);
 	  cmdmode = false;
 		return;
 	}
 	switch(cmd[1])
 	{
 		case GETDATA : read_EE(); break;
 		case SETTIME : 
 			if(cc == 7)
 			{
 				printf(lcd_putc,"\fStelle die\nUhr");
 				hh    = cmd[2];
 				mm    = cmd[3];
 				ss    = cmd[4];
 				month = cmd[5];
 				day   = cmd[6];
 				oldhh = hh;
 				WriteTime();
 			}
 			break;
 		case RESET :
 			clear_EE();
 			break;
 		case RESTART:
 			if(cc == 7)
 			{
 				printf(lcd_putc,"\fStelle die\nUhr");
 				hh    = cmd[2];
 				mm    = cmd[3];
 				ss    = cmd[4];
 				month = cmd[5];
 				day   = cmd[6];
 				oldhh = hh;
 				WriteTime();
 				delay_ms(1000);
 				printf(lcd_putc,"\fRESTART");
 		  	index = 0x02;
 		 		write_eeprom(0x00,month);
 				write_eeprom(0x01,day);
 				delay_ms(1000);
 			}
 			break;
 		default      : break;
 	} 
 	disable_interrupts(int_ext);
 	cmdmode = false;		
}


/**************************************************************************/
/*                                                                        */
/**************************************************************************/
void ReadTime(void)
{
  i2c_Start();               // Set a START condition I2C Bus
  i2c_Write( 0b10100010 );   // Address and Write Flag
  i2c_Write( 0b00000010 );   // read from Byte 2
  i2c_Start();               // Set a START condition I2C Bus
  i2c_Write( 0b10100011 );   // Address and Read Flag
  ss = BcdToChar(i2c_Read());  // sec
  mm = BcdToChar(i2c_Read());  // min
  hh = BcdToChar(i2c_Read(0)); // hours and NO_ACK;
  i2c_stop();
}


/**************************************************************************/
/*                                                                        */
/**************************************************************************/
void WriteTime(void)
{
  i2c_Start();                  // Set a START condition I2C Bus
  i2c_Write( 0b10100010 );      // Address and Write Flag
  i2c_Write( 0b00000010 );      // write from Byte 2
  i2c_Write(CharToBcd(ss));     // sec
  i2c_Write(CharToBcd(mm));     // min
  i2c_Write(CharToBcd(hh));     // hh
  i2c_stop();
}



/**************************************************************************/
/*                                                                        */
/**************************************************************************/
char BcdToChar(char c)
{
   return ((c/16)*10 + (c%16));
}




/**************************************************************************/
/*                                                                        */
/**************************************************************************/
char CharToBcd(char b)
{
   return((b/10)*16 + (b%10));
}


/**************************************************************************/
/*                                                                        */
/**************************************************************************/
#use rs232(baud=PCBAUD,xmit=PIN_B1, rcv=PIN_B0, BITS=8)
read_EE()
{
	byte adr;
	putc(index);
	for(adr = 0; adr <= 253; adr++)
		putc(read_eeprom(adr));
}



clear_EE()
{
	byte adr;
	adr = 0;
	printf(lcd_putc,"\fLoesche EEProm");
	do
		write_eeprom(adr++,0x00);
	while(adr != 254);
}